package com.xyzc.start.cache;

import cn.hutool.core.collection.CollUtil;
import com.xyzc.common.redis.cache.RedisCache;
import com.xyzc.start.domain.AgentConfig;
import com.xyzc.start.domain.WxConfig;
import lombok.RequiredArgsConstructor;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

@Service
@RequiredArgsConstructor
public class SysParamsCache {

    private final RedisCache redisCache;

    /**
     * 参数管理 KEY
     */
    private final String SYSTEM_PARAMS_KEY = "agent:weixin:appId:";

    /**
     * 微信用户登录状态管理 KEY
     */
    private final String WX_USER_LOGIN_STATUS_KEY = "wx:userLogin:status:";

    /**
     * 保存参数列表到缓存
     *
     * @param list 参数列表
     */
    public void saveList(List<WxConfig> list) {
        if (CollUtil.isEmpty(list)) {
            return;
        }
        // list 转成 map
        //如果merId为空，则跳过，不收集
        List<WxConfig> collect = list.stream().filter(wxConfig -> wxConfig.getMerId() != null).toList();

        Map<String, Object> map = collect.stream().collect(Collectors.toMap(WxConfig::getMerId, Function.identity()));

        redisCache.hMSet(SYSTEM_PARAMS_KEY, map, RedisCache.NOT_EXPIRE);
    }

    /**
     * 删除缓存中的全部参数列表
     */
    public void delList() {
        redisCache.delete(SYSTEM_PARAMS_KEY);
    }

    /**
     * 根据参数键，获取参数值
     *
     * @param paramKey 参数键
     */
    public String get(String paramKey) {
        return (String) redisCache.hGet(SYSTEM_PARAMS_KEY, paramKey);
    }

    /**
     *
     */
    public AgentConfig getAgentConfig(String agentMerId) {
        return (AgentConfig) redisCache.hGet(SYSTEM_PARAMS_KEY, agentMerId);
    }

    /**
     * 根据参数键，获取参数值
     *
     * @param paramKey   参数键
     * @param paramValue 参数值
     */
    public void save(String paramKey, String paramValue) {
        redisCache.hSet(SYSTEM_PARAMS_KEY, paramKey, paramValue);
    }

    /**
     * 根据参数键，删除参数值
     *
     * @param paramKey 参数键
     */
    public void del(String... paramKey) {
        redisCache.hDel(SYSTEM_PARAMS_KEY, paramKey);
    }

    public void clear() {
        redisCache.hDel(SYSTEM_PARAMS_KEY, "*");
    }

    /**
     * 设置微信用户登录状态
     */
    public void setUserStatus(String ticket, String openid) {
        redisCache.set(WX_USER_LOGIN_STATUS_KEY + ticket, openid, RedisCache.MINUTE_FIVE_EXPIRE);
    }

    /**
     * 使用 ticket 获取微信用户登录状态
     */
    public String getUserStatus(String ticket) {
        return (String) redisCache.get(WX_USER_LOGIN_STATUS_KEY + ticket);
    }

    /**
     * 删除微信用户登录状态
     *
     * @param openId
     */
    public void delUserStatus(String openId) {
        redisCache.delete(WX_USER_LOGIN_STATUS_KEY + openId);
    }
}
