package com.xyzc.start.service.impl;

import com.xyzc.common.core.exception.ServerException;
import com.xyzc.start.domain.DNSVO;
import com.xyzc.start.domain.bo.MerchantDNSBo;
import com.xyzc.start.mapper.MerchantMapper;
import com.xyzc.start.service.IDNSService;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.xbill.DNS.Record;
import org.xbill.DNS.*;

import java.io.IOException;
import java.util.List;

@Service
@RequiredArgsConstructor
@Slf4j
public class DNSServiceImpl implements IDNSService {

    private final MerchantMapper merchantMapper;

    /**
     * 小易给商户开发的 三级域名泛解析，改动时需要配合域名泛解析一起改动
     */
    private final String PRIVATE_HOST = ".xy.xiaoyizc.com";

    /**
     * 用于域名验证的 txt 记录，让商户自己去后台配置域名验证
     */
    private final String TXT_RECORD_VALUE = "da31cg3kjs23z34hfs284a72sa12";

    private final String TXT_RECORD_PREFIX = "verify.";

    private final  String  suffixName = ".wangwu.itfenxiang.cn";


    @Override
    public DNSVO checkDomain(MerchantDNSBo bo) {
        //查询代理商填写的私域，然后获取txt 值进行验证是否是代理商本人的私域


        String privateDomain = bo.getPrivateDomain();
        privateDomain = TXT_RECORD_PREFIX + privateDomain;
        try {
            // 查询TXT记录
            List<TXTRecord> txtRecords = lookupTXTRecords(privateDomain);
            for (TXTRecord record : txtRecords) {
                //txt 记录为空，抛出
                if (record.getStrings().isEmpty()) {
                    throw new ServerException("TXT Record is empty");
                }

                //txt 记录不为空，但是值不正确，抛出
                if (!TXT_RECORD_VALUE.equals(record.getStrings().get(0))) {
                    throw new ServerException("无法绑定域名，域名的空主机头（TXT）记录与预设不符。如果已经设置，可能是暂时没有生效，请稍候再试。" + privateDomain);
                }
            }
        } catch (IOException e) {
            log.info("Failed to lookup TXT records for " + privateDomain + ": " + e.getMessage());
        }

        //没问题就封装返回商户独立的 cname
        DNSVO vo = new DNSVO();
        vo.setCname(bo.getPrivateDomain() + PRIVATE_HOST);
        vo.setPrivateDomain(bo.getPrivateDomain());
        vo.setDomainPrefix(bo.getDomainPrefix());
        return vo;
    }


    /**
     * 根据域名查询TXT记录
     *
     * @param domain
     * @return
     * @throws IOException
     */
    private static List<TXTRecord> lookupTXTRecords(String domain) throws IOException {

        // 创建一个DNS查询对象
        Lookup lookup = new Lookup(domain, Type.TXT);
        // 设置DNS服务器
        lookup.setResolver(new SimpleResolver("114.114.114.114"));
        // 设置超时时间
        Record[] records = lookup.run();
        //
        if (lookup.getResult() == Lookup.SUCCESSFUL) {
            List<TXTRecord> txtRecords = new java.util.ArrayList<>();
            for (Record record : records) {
                if (record instanceof TXTRecord) {
                    txtRecords.add((TXTRecord) record);
                }
            }
            return txtRecords;
        } else {
            throw new IOException("Failed to lookup TXT records for " + domain + ": " + lookup.getErrorString());
        }
    }

}
