package com.xyzc.start.service.impl;

import cn.hutool.core.util.RandomUtil;
import com.xyzc.common.core.exception.ErrorCode;
import com.xyzc.common.core.exception.ServerException;
import com.xyzc.common.core.utils.AssertUtils;
import com.xyzc.common.sms.cache.SmsSendCache;
import com.xyzc.common.sms.service.ISmsService;
import com.xyzc.common.web.security.mobile.MobileAuthenticationToken;
import com.xyzc.common.web.security.user.UserDetail;
import com.xyzc.common.web.security.cache.TokenStoreCache;
import com.xyzc.start.constant.BaseUserConst;
import com.xyzc.start.convert.BaseUserConvert;
import com.xyzc.start.domain.BaseUser;
import com.xyzc.start.domain.bo.BaseUserLoginBo;
import com.xyzc.start.domain.bo.UserMobileLoginBO;
import com.xyzc.start.domain.vo.AccessTokenVo;
import com.xyzc.start.domain.vo.BaseUserMemberCardVo;
import com.xyzc.start.domain.vo.SysOssVo;
import com.xyzc.start.mapper.BaseUserMapper;
import com.xyzc.start.mapper.SysOssMapper;
import com.xyzc.start.service.ICaptchaService;
import com.xyzc.start.service.IUserAuthService;
import com.xyzc.common.core.utils.TokenUtils;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.util.List;


/**
 * 用户登录
 *
 * @author 墨凡
 * date: 2024/7/21 23:27
 */
@Service
@RequiredArgsConstructor
@Slf4j
public class UserAuthServiceImpl implements IUserAuthService {

    private final TokenStoreCache tokenStoreCache;
    private final AuthenticationManager authenticationManager;

    private final ISmsService smsService;
    private final SmsSendCache smsSendCache;
    private final ICaptchaService captchaService;

    private final BaseUserMapper baseUserMapper;
    private final SysOssMapper sysOssMapper;


    @Override
    @Transactional(rollbackFor = Exception.class)
    public boolean sendCode(UserMobileLoginBO login) {
        //todo 后续删除
        if (login != null) {
            return true;
        }

        // 验证图形验证码
        boolean validate = captchaService.validate(login.getKey(), login.getCode());
        AssertUtils.isTrue(validate, ErrorCode.CAPTCHA_ERROR);

        // TODO :短信是否发送达到上限

        // 如果数据库中还有短信验证码就先不发送验证码了，直接返回错误
        String code = smsSendCache.getCode(login.getMobile());
        AssertUtils.isNull(code, ErrorCode.SEND_MOBILE_FAST);

        // 生成6位验证码
        code = RandomUtil.randomNumbers(6);

        // TODO： 后期改为正式代理商
        BaseUser user = baseUserMapper.queryDetailByMobile(login.getMobile(), 1810310843224322049L);

        // 发送短信
        boolean sendFlag = smsService.sendCode(login.getMobile(), "code", code);

        // TODO： 后期新添加用户抽离出来不要放在这里
        // 未发现用户就保存用户信息
        if (user == null) {
            BaseUser baseUser = new BaseUser();
            baseUser.setUserName(login.getMobile());
            baseUser.setMobile(login.getMobile());
            baseUser.setStatus(BaseUserConst.NORMAL_STATUS);
            baseUser.setSurplusFuel(10L);

            //TODO ：代理商id
            baseUser.setAgentMerId(1810310843224322049L);
            baseUserMapper.insert(baseUser);
        }

        return sendFlag;
    }

    @Override
    public AccessTokenVo loginByAccount(BaseUserLoginBo login) {
        Authentication authentication;
        try {
            // 用户认证
            authentication = authenticationManager.authenticate(
                    new UsernamePasswordAuthenticationToken(login.getUsername(), login.getPassword()));
        } catch (BadCredentialsException e) {
            throw new ServerException("用户名或密码错误");
        }

        // 用户信息
        UserDetail user = (UserDetail) authentication.getPrincipal();

        // 生成 accessToken
        String accessToken = TokenUtils.generator();

        // 保存用户信息到缓存
        tokenStoreCache.saveUser(accessToken, user);

        AccessTokenVo baseUserTokenVo = new AccessTokenVo();
        baseUserTokenVo.setAccess_token(accessToken);

        return baseUserTokenVo;
    }

    @Override
    public void logout(String accessToken) {
        // 用户信息
        UserDetail user = tokenStoreCache.getUser(accessToken);

        // 删除用户信息
        tokenStoreCache.deleteUser(accessToken);

        // TODO: 保存登录日志
    }

    /**
     * 注册新用户
     *
     * @param register BaseUserLoginBo
     */
    @Override
    public void register(BaseUserLoginBo register) {
        // TODO: 注册用户

    }

    /**
     * 手机号登录
     *
     * @param login UserMobileLoginBO
     * @return BaseUserTokenVo
     */
    @Override
    public AccessTokenVo loginByMobile(UserMobileLoginBO login) {
        Authentication authentication;
        try {
            // 用户认证
            authentication = authenticationManager.authenticate(
                    new MobileAuthenticationToken(login.getMobile(), login.getCode()));
        } catch (BadCredentialsException e) {
            throw new ServerException(ErrorCode.SEND_MOBILE_ERROR);
        }

        // 用户信息
        UserDetail userDetail = (UserDetail) authentication.getPrincipal();

        // 生成 accessToken
        String accessToken = TokenUtils.generator();

        // 保存用户信息到缓存
        tokenStoreCache.saveUser(accessToken, userDetail);

        // 转换为登录用户信息
        AccessTokenVo baseUserTokenVo = BaseUserConvert.INSTANCE.convertTokenVo(userDetail);
        baseUserTokenVo.setAccess_token(accessToken);
        baseUserTokenVo.setUserName(userDetail.getUsername());
        //
        SysOssVo sysOssVo = sysOssMapper.selectVoById(userDetail.getAvatar());
        baseUserTokenVo.setAvatar(sysOssVo.getUrl());
        // TODO 剩余次数，虚拟
        baseUserTokenVo.setUseCount(109);

        // 添加用户拥有的会员卡
        List<BaseUserMemberCardVo> memberCardVos = baseUserMapper.queryMyMemberCardByUserId(userDetail.getId());
        baseUserTokenVo.setMemberCardVoList(memberCardVos);

        // 添加用户会员卡数量
        baseUserTokenVo.setMemberCardCount(memberCardVos.size());

        return baseUserTokenVo;
    }
}
