package com.xyzc.start.service.impl;

import com.xyzc.common.core.exception.ErrorCode;
import com.xyzc.common.core.exception.ServerException;
import com.xyzc.common.core.utils.AssertUtils;
import com.xyzc.common.core.utils.StringUtils;
import com.xyzc.common.core.utils.TokenUtils;
import com.xyzc.common.web.security.cache.TokenStoreCache;
import com.xyzc.common.web.security.mobile.WeiXinScanAuthenticationToken;
import com.xyzc.common.web.security.user.UserDetail;
import com.xyzc.start.cache.SysParamsCache;
import com.xyzc.start.domain.AgentConfig;
import com.xyzc.start.domain.BaseUser;
import com.xyzc.start.domain.WxConfig;
import com.xyzc.start.domain.vo.AccessTokenVo;
import com.xyzc.start.domain.vo.TicketVo;
import com.xyzc.start.mapper.BaseUserMapper;
import com.xyzc.start.mapper.WxConfigMapper;
import com.xyzc.start.service.IWeiXinAuthService;
import jakarta.annotation.PostConstruct;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import me.chanjar.weixin.common.bean.oauth2.WxOAuth2AccessToken;
import me.chanjar.weixin.common.error.WxErrorException;
import me.chanjar.weixin.mp.api.WxMpService;
import me.chanjar.weixin.mp.bean.result.WxMpUser;
import me.chanjar.weixin.mp.config.impl.WxMpDefaultConfigImpl;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.util.EntityUtils;
import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;
import org.springframework.boot.configurationprocessor.json.JSONObject;
import org.springframework.security.authentication.AuthenticationManager;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.core.Authentication;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.io.IOException;
import java.io.InputStream;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

@Service
@RequiredArgsConstructor
@Slf4j
public class WeiXinAuthServiceImpl implements IWeiXinAuthService {

    private final TokenStoreCache tokenStoreCache;
    private final AuthenticationManager authenticationManager;
    private final WxConfigMapper wxConfigMapper;
    private final SysParamsCache sysParamsCache;
    private final WxMpService wxMpService;
    private final BaseUserMapper baseUserMapper;

    private static final String APPID = "wxb3a7748bbf94e108";
    private static final String APPSECRET = "7e43a7f547cb150d6c86b4b66c473daf";

    /**
     * 启动项目时，初始化微信公众号的配置
     */
    @PostConstruct
    public void init() {
        // 从数据库读取所有的微信公众号配置
        List<WxConfig> wxConfigVoList = wxConfigMapper.selectList(null);

        // 将上面的数据按照上下存储到redis
        sysParamsCache.saveList(wxConfigVoList);
    }

    /**
     * 微信扫码登录
     *
     * @param agentId Long
     * @return
     */
    @Override
    public TicketVo sacn(Long agentId) {
        // 获取商户信息，商户没有微信公众号，就返回错误信息
        AgentConfig agentConfig = sysParamsCache.getAgentConfig(String.valueOf(agentId));
        AssertUtils.notNull(agentConfig, ErrorCode.COMPANY_NO_OPEN_WEIXIN);
        try {
            // sceneStr 是扫码后微信返回的参数
            return generateQRCode("wx:sceneStr");
        } catch (Exception e) {
            throw new ServerException(ErrorCode.COMPANY_NO_OPEN_WEIXIN);
        }
    }

    /**
     * 获取access_token
     *
     * @return
     * @throws Exception
     */
    public static String getAccessToken() throws Exception {
        HttpClient httpClient = HttpClients.createDefault();
        // 拼接获取 access_token 的 url
        String url = "https://api.weixin.qq.com/cgi-bin/token?" +
                "grant_type=client_credential" +
                "&appid=" + APPID +
                "&secret=" + APPSECRET;
        HttpGet httpGet = new HttpGet(url);
        HttpResponse httpResponse = httpClient.execute(httpGet);
        String response = EntityUtils.toString(httpResponse.getEntity());
        JSONObject jsonObject = new JSONObject(response);
        return jsonObject.getString("access_token");
    }


    /**
     * 获取二维码票据
     *
     * @param sceneStr String
     * @return TicketVo
     */
    public static TicketVo generateQRCode(String sceneStr) throws Exception {
        HttpClient httpClient = HttpClients.createDefault();
        String accessToken = getAccessToken();
        String url = "https://api.weixin.qq.com/cgi-bin/qrcode/create?access_token=" + accessToken;
        String postData = "{\"expire_seconds\": 1800, \"action_name\": \"QR_STR_SCENE\", \"action_info\": {\"scene\": {\"scene_str\": \"" + sceneStr + "\"}}}";
        HttpPost httpPost = new HttpPost(url);
        httpPost.setEntity(new StringEntity(postData, "UTF-8"));
        HttpResponse httpResponse = httpClient.execute(httpPost);
        String response = EntityUtils.toString(httpResponse.getEntity());

        // 获取ticket
        org.springframework.boot.configurationprocessor.json.JSONObject jsonObject = new org.springframework.boot.configurationprocessor.json.JSONObject(response);
        String ticket = jsonObject.getString("ticket");
        TicketVo ticketVo = new TicketVo();
        ticketVo.setTicket(ticket);
        ticketVo.setCodeUrl("https://mp.weixin.qq.com/cgi-bin/showqrcode?ticket=" + URLEncoder.encode(ticket, "UTF-8"));
        return ticketVo;
    }


    /**
     * 微信网页授权登录
     *
     * @param code
     * @param agentId
     * @return
     */
    @Override
    public AccessTokenVo auth(String code, Long agentId) {
        // 帮我生成微信授权携带的code码
        getBaseUserInfoByOpenid(code, agentId);

        return null;
    }

    @Override
    public void postToken(HttpServletRequest req, HttpServletResponse res) {
        final Map<String, String> map = parseXml(req);

        //用户openid
        final String fromUserName = map.get("FromUserName");
        //消息类型，event
        final String msgType = map.get("MsgType");
        //事件类型
        final String event = map.get("Event");
        // ticket
        final String ticket = map.get("Ticket");

        String scene_str = "";
        if ("event".equals(msgType)) {
            if (event.equals("subscribe")) {
                if (ticket != null) {
                    scene_str = map.get("EventKey").replace("qrscene_", "");
                }
            }
            // 事件类型为SCAN即已关注
            else if (event.equals("SCAN")) {
                if (ticket != null) {
                    scene_str = map.get("EventKey");
                }
            }
        } else if ("text".equals(msgType)) {
            // 事件类型为text，用户发送的聊天消息
            final String content = map.get("Content");
            log.info("用户发送的消息：{}", content);
        }
        // 如果scene_str 不为空代表用户已经扫描并且关注了公众号
        if (StringUtils.isNotEmpty(scene_str)) {
            // 将微信公众号用户ID缓存到redis中，标记用户已经扫码完成，执行登录逻辑。
            sysParamsCache.setUserStatus(ticket, fromUserName);
        }
        //返回消息
        try {
            res.getWriter().write("success");
        } catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 扫码循环判断是否登录
     * @param ticket
     * @return
     */
    @Override
    @Transactional(rollbackFor = Exception.class)
    public AccessTokenVo getStatus(String ticket) {
        String userOpenid = sysParamsCache.getUserStatus(ticket);
        if (userOpenid == null || userOpenid.isEmpty()) {
            throw new ServerException(ErrorCode.NOT_SCAN_CODE);
        }

        // 获取用户信息
        try {
            Authentication authentication;
            // 用户认证
            authentication = authenticationManager.authenticate(new WeiXinScanAuthenticationToken(userOpenid, 1810310843224322049L));
            UserDetail user = (UserDetail) authentication.getPrincipal();
            // 生成 accessToken
            String accessToken = TokenUtils.generator();

            // 保存用户信息到缓存
            tokenStoreCache.saveUser(accessToken, user);

            AccessTokenVo baseUserTokenVo = new AccessTokenVo();
            baseUserTokenVo.setAccess_token(accessToken);

            // 清除ticket
            sysParamsCache.delUserStatus(ticket);
            return baseUserTokenVo;
        } catch (BadCredentialsException e) {
            throw new ServerException(ErrorCode.QR_CODE_ERROR);
        }
    }

    /**
     * 解析xml数据
     *
     * @param req
     * @return
     */
    private Map<String, String> parseXml(HttpServletRequest req) {
        Map<String, String> map = new HashMap<>();
        InputStream inputStream = null;
        try {
            inputStream = req.getInputStream();
            SAXReader reader = new SAXReader();
            Document document = reader.read(inputStream);
            Element root = document.getRootElement();
            List<Element> list = root.elements();
            for (Element e : list) {
                map.put(e.getName(), e.getText());
            }
        } catch (Exception e) {
            throw new RuntimeException(e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    throw new RuntimeException(e);
                }
            }
        }
        return map;
    }

    private AccessTokenVo getBaseUserInfoByOpenid(String code, Long agentId) {
        try {
            String appId = "wxb3a7748bbf94e108";
            String secret = "7e43a7f547cb150d6c86b4b66c473daf";
            WxMpDefaultConfigImpl wxMpDefaultConfig = new WxMpDefaultConfigImpl();
            wxMpDefaultConfig.setAppId(appId);
            wxMpDefaultConfig.setSecret(secret);
            wxMpService.setWxMpConfigStorage(wxMpDefaultConfig);
//            wxMpService.switchover("wxb3a7748bbf94e108");
            // 根据openid查询用户
            WxOAuth2AccessToken accessToken;
            accessToken = wxMpService.getOAuth2Service().getAccessToken(code);
            String openId = accessToken.getOpenId();
            log.info("微信授权成功，openId:{}", openId);
            //根据 openId 查询用户是否授权登录过
            BaseUser baseUser = baseUserMapper.queryUserByOpenid(openId, agentId);
            log.info("开始转换用户,{}", baseUser);

            if (!Objects.isNull(baseUser)) {
                //已经授权登录过，更新redis的token，返回用户信息
                AccessTokenVo baseUserTokenVo = new AccessTokenVo();
                return baseUserTokenVo;
            } else {
                // 未授权登录过，授权，登记用户信息
                WxMpUser wxMpUser = wxMpService.getUserService().userInfo(accessToken.getOpenId());
                baseUser.setAgentMerId(agentId);
                //新增一个系统用户，不给这个用户账号密码
                //绑定微信基础用户角色
//                SysRole sysRole = roleMapper.selectRoleIdByRoleKey(MerTypeConst.BASE_USER);
//                roleService.insertAuthUsers(sysRole.getRoleId(), new Long[]{sysUserBo.getUserId()});
                AccessTokenVo baseUserTokenVo = new AccessTokenVo();
                return baseUserTokenVo;
            }
        } catch (WxErrorException e) {
            log.info("微信授权失败，错误信息: {}", e.getMessage());
            throw new ServerException("微信授权失败，错误信息：" + e.getMessage());
        }
    }

}
